# Laden der benoetigten Pakete:
library(checkmate)
library(ggplot2)
library(ggrepel)
library(rgl)

# Zielfunktionen:
f1 = function(x) (x + 3)^2
f2 = function(x) x^2 - 5 * cos(10 * x)
f3 = function(x) x[1]^2 + x[2]^2
f4 = function(x) x[1] * sin(x[1]) + 3 * x[2]^2

# Funktion, um das Ergebnis der Kompasssuche bei einer eindimensionalen Funktion
# zu plotten:
plot_f_1d = function(f, res_kompass, lower, upper, title = "") {
  # check input:
  assert_function(x = f, args = "x")
  assert_data_frame(x = res_kompass, ncols = 3)
  assert_names(names(x = res_kompass), permutation.of = c("s", "x1", "fx")) 
  assert_number(x = lower, finite = TRUE)
  assert_number(x = upper, lower = lower, finite = TRUE)

  print(ggplot(data = res_kompass, mapping = aes(x = x1, y = fx)) +
          geom_point() +
          stat_function(fun = f, xlim = c(lower, upper), n = 1001) +
          geom_text_repel(mapping = aes(x = x1, y = fx, 
                                        label = 0:(nrow(res_kompass) - 1))) +
          labs(title = title, x = "x", y = "f(x)") +
          theme(plot.title = element_text(hjust = 0.5),
                legend.key = element_rect(fill = "transparent", colour = NA),
                legend.title = element_blank()))
}

# Funktion, um das Ergebnis der Kompasssuche bei einer zweidimensionalen Funktion
# zu plotten:
plot_f_2d = function(f, res_kompass, lower, upper, title = "") {
  # check input:
  assert_function(f, args = "x")
  assert_data_frame(res_kompass, ncols = 4)
  assert_names(names(res_kompass), permutation.of = c("s", "x1", "x2", "fx")) 
  assert_numeric(lower, len = 2, finite = TRUE)
  assert_numeric(upper, len = 2, finite = TRUE)

  # plot function:
  grid = expand.grid(x = seq(lower[1], upper[1], length.out = 51),
                     y = seq(lower[2], upper[2], length.out = 51))
  data = data.frame(grid, z = apply(grid, 1, f))
  plot3d(x = data$x, y = data$y, z = data$z, xlab = "x", ylab = "y",
         zlab = "f(x, y)", main = title)

  # plot steps of optimization:
  points3d(x = res_kompass$x1, y = res_kompass$x2, z = res_kompass$fx,
           col = "red", size = 5)
  text3d(x = res_kompass$x1, y = res_kompass$x2, z = res_kompass$fx,
         texts = 0:(nrow(res_kompass) - 1))
}

# Funktion, die erst die Kompasssuche ausfueht und anschliessend das Ergebnis
# plottet (falls plot = TRUE):
## Eingabe:
# f: Funktion, die optimiert werden soll
# x0: initialer Wert bzw. initiale Werte fuer x
# s0: initialer Wert fuer s, sollte > 0 sein
# theta: initialer Wert fuer theta, sollte aus (0, 1) sein
# lower: untere Grenze fuer x bzw. x und y
# upper: ober Grenze fuer x bzw. x und y
# iters: Anzahl der Iterationen der Kompasssuche
# title: Titel der Grafik
# plot: TRUE fuer plotten, FALSE fuer nicht plotten
# nrows: Anzahl der Zeilen des Ergebnisses, die geplottet werden sollen
## Ausgabe:
# Ergebnis der Kompasssuche
kompass_plot_f = function(f, x0, s0, theta, lower, upper, iters, title = "",
                          plot = TRUE, nrows = NULL) {
  # check input:
  assert_function(f, args = "x")
  assert_numeric(x0, min.len = 1, max.len = 2, finite = TRUE)
  assert_number(s0, lower = 0, finite = TRUE)
  assert_number(theta, lower = 0, upper = 1)
  assert_numeric(lower, min.len = 1, max.len = 2, finite = TRUE)
  assert_numeric(upper, min.len = 1, max.len = 2, finite = TRUE)
  assert_count(iters, positive = TRUE)
  assert_flag(plot)

  # perform optimization:
  res_kompass = kompasssuche(f = f, x0 = x0, s0 = s0, theta = theta, iters = iters)
  assert_number(nrows, lower = 1, upper = nrow(res_kompass), null.ok = TRUE)
  if(is.null(nrows)) nrows = nrow(res_kompass)

  # plot results:
  if(plot) {
    if(length(x0) == 1) {
      plot_f_1d(f = f, res_kompass = res_kompass[1:nrows, ], lower = lower,
                upper = upper, title = title)
    } else if(length(x0 == 2)) {
      plot_f_2d(f = f, res_kompass = res_kompass[1:nrows, ], lower = lower,
                upper = upper, title = title)
    }
  }
  return(res_kompass)
}
